# frozen_string_literal: true

require 'rubocop_spec_helper'
require_relative '../../../../rubocop/cop/background_migration/dictionary_file'

RSpec.describe RuboCop::Cop::BackgroundMigration::DictionaryFile, feature_category: :database do
  let(:config) do
    RuboCop::Config.new(
      'BackgroundMigration/DictionaryFile' => {
        'EnforcedSince' => 20231018100907
      }
    )
  end

  shared_examples 'migration with missing dictionary keys offense' do |missing_key|
    let(:missing_key_message) do
      format(described_class::MSG[:missing_key], key: missing_key, filename: dictionary_file_path)
    end

    it 'registers an offense' do
      expect_offense(<<~RUBY)
        class QueueMyMigration < Gitlab::Database::Migration[2.1]
        ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ #{missing_key_message}
          MIGRATION = 'MyMigration'

          def up
            queue_batched_background_migration(
              MIGRATION,
              :users,
              :id
            )
          end
        end
      RUBY
    end
  end

  context 'for non post migrations' do
    before do
      allow(cop).to receive(:in_post_deployment_migration?).and_return(false)
    end

    it 'does not throw any offense' do
      expect_no_offenses(<<~RUBY)
        class QueueMyMigration < Gitlab::Database::Migration[2.1]
          MIGRATION = 'MyMigration'

          def up
            queue_batched_background_migration(
              MIGRATION,
              :users,
              :id
            )
          end
        end
      RUBY
    end
  end

  context 'for post migrations' do
    before do
      allow(cop).to receive(:in_post_deployment_migration?).and_return(true)
    end

    context 'without enqueuing batched migrations' do
      it 'does not throw any offense' do
        expect_no_offenses(<<~RUBY)
          class CreateTestTable < Gitlab::Database::Migration[2.1]
            def change
              create_table(:tests)
            end
          end
        RUBY
      end
    end

    context 'with enqueuing batched migration' do
      let(:rails_root) { File.expand_path('../../../..', __dir__) }
      let(:dictionary_file_path) { File.join(rails_root, 'db/docs/batched_background_migrations/my_migration.yml') }

      context 'for migrations before enforced time' do
        before do
          allow(cop).to receive(:version).and_return(20230918100907)
        end

        it 'does not throw any offenses' do
          expect_no_offenses(<<~RUBY)
            class QueueMyMigration < Gitlab::Database::Migration[2.1]
              MIGRATION = 'MyMigration'

              def up
                queue_batched_background_migration(
                  MIGRATION,
                  :users,
                  :id
                )
              end
            end
          RUBY
        end
      end

      context 'for migrations after enforced time' do
        let(:missing_dictionary_message) do
          format(described_class::MSG[:missing_dictionary], filename: dictionary_file_path)
        end

        before do
          allow(cop).to receive(:version).and_return(20231118100907)
        end

        it 'throws offense on not having the appropriate dictionary file with migration name as a constant' do
          expect_offense(<<~RUBY)
            class QueueMyMigration < Gitlab::Database::Migration[2.1]
            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ #{missing_dictionary_message}
              MIGRATION = 'MyMigration'

              def up
                queue_batched_background_migration(
                  MIGRATION,
                  :users,
                  :id
                )
              end
            end
          RUBY
        end

        it 'throws offense on not having the appropriate dictionary file with migration name as a variable' do
          expect_offense(<<~RUBY)
            class QueueMyMigration < Gitlab::Database::Migration[2.1]
            ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ #{missing_dictionary_message}
              def up
                queue_batched_background_migration(
                  'MyMigration',
                  :users,
                  :id
                )
              end
            end
          RUBY
        end

        context 'with dictionary file' do
          let(:introduced_by_url) { 'https://gitlab.com/gitlab-org/gitlab/-/merge_requests/132639' }
          let(:milestone) { '16.1' }

          before do
            allow(File).to receive(:exist?).and_call_original
            allow(File).to receive(:exist?).with(dictionary_file_path).and_return(true)
            allow(Gitlab::Utils::BatchedBackgroundMigrationsDictionary)
              .to receive(:entries).and_return({
                '20231118100907' => {
                  introduced_by_url: introduced_by_url,
                  milestone: milestone
                }
              })
          end

          context 'without introduced_by_url' do
            it_behaves_like 'migration with missing dictionary keys offense', :introduced_by_url do
              let(:introduced_by_url) { nil }
            end
          end

          context 'when the `introduced_by_url` is not correct' do
            let(:introduced_by_url) { 'https://gitlab.com/gitlab-org/gitlab/-/merge_requests/132639/invalid' }
            let(:invalid_url_message) do
              format(described_class::MSG[:invalid_url], key: :introduced_by_url, filename: dictionary_file_path)
            end

            it 'throws offense on having a correct url' do
              expect_offense(<<~RUBY)
                class QueueMyMigration < Gitlab::Database::Migration[2.1]
                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ #{invalid_url_message}
                  def up
                    queue_batched_background_migration(
                      'MyMigration',
                      :users,
                      :id
                    )
                  end
                end
              RUBY
            end
          end

          context 'without milestone' do
            it_behaves_like 'migration with missing dictionary keys offense', :milestone do
              let(:milestone) { nil }
            end
          end

          context 'when milestone is a number' do
            let(:milestone) { 16.1 }
            let(:invalid_milestone_message) do
              format(described_class::MSG[:invalid_milestone], key: :milestone, filename: dictionary_file_path)
            end

            it 'throws offense on having an invalid milestone' do
              expect_offense(<<~RUBY)
                class QueueMyMigration < Gitlab::Database::Migration[2.1]
                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ #{invalid_milestone_message}
                  def up
                    queue_batched_background_migration(
                      'MyMigration',
                      :users,
                      :id
                    )
                  end
                end
              RUBY
            end
          end

          context 'with required dictionary keys' do
            it 'does not throw offense with appropriate dictionary file' do
              expect_no_offenses(<<~RUBY)
                class QueueMyMigration < Gitlab::Database::Migration[2.1]
                  MIGRATION = 'MyMigration'

                  def up
                    queue_batched_background_migration(
                      MIGRATION,
                      :users,
                      :id
                    )
                  end
                end
              RUBY
            end
          end
        end
      end
    end

    context 'with ensure_batched_background_migration_is_finished' do
      let(:rails_root) { File.expand_path('../../../..', __dir__) }
      let(:dictionary_file_path) { File.join(rails_root, 'db/docs/batched_background_migrations/my_migration.yml') }

      context 'for migrations before enforced time' do
        before do
          allow(cop).to receive(:version).and_return(20230918100907)
        end

        it 'does not throw any offenses' do
          expect_no_offenses(<<~RUBY)
            class FinalizeMyMigration < Gitlab::Database::Migration[2.1]
              MIGRATION = 'MyMigration'

              def up
                ensure_batched_background_migration_is_finished(
                  job_class_name: MIGRATION,
                  table_name: :users,
                  column_name: :id,
                  job_arguments: []
                )
              end
            end
          RUBY
        end
      end

      context 'for migrations after enforced time' do
        before do
          allow(cop).to receive(:version).and_return(20231118100907)
        end

        context 'with dictionary file' do
          let(:introduced_by_url) { 'https://gitlab.com/gitlab-org/gitlab/-/merge_requests/132639' }
          let(:milestone) { '16.1' }
          let(:finalized_by) { '20231118100907' }

          before do
            dictionary_instance = instance_double(Gitlab::Utils::BatchedBackgroundMigrationsDictionary)
            allow(dictionary_instance).to receive(:finalized_by).and_return(finalized_by)

            allow(Gitlab::Utils::BatchedBackgroundMigrationsDictionary)
              .to receive(:entry).with('MyMigration').and_return(dictionary_instance)
          end

          context 'without finalized_by' do
            let(:finalized_by) { nil }
            let(:missing_finalized_by_message) do
              format(described_class::MSG[:missing_finalized_by], filename: dictionary_file_path)
            end

            it 'throws offense on missing finalized_by' do
              expect_offense(<<~RUBY)
                class FinalizeMyMigration < Gitlab::Database::Migration[2.1]
                ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ #{missing_finalized_by_message}
                  MIGRATION = 'MyMigration'

                  def up
                    ensure_batched_background_migration_is_finished(
                      job_class_name: MIGRATION,
                      table_name: :users,
                      column_name: :id,
                      job_arguments: []
                    )
                  end
                end
              RUBY
            end
          end

          context 'with finalized_by' do
            it 'does not throw offense with appropriate dictionary file' do
              expect_no_offenses(<<~RUBY)
                class FinalizeMyMigration < Gitlab::Database::Migration[2.1]
                  MIGRATION = 'MyMigration'

                  def up
                    ensure_batched_background_migration_is_finished(
                      job_class_name: MIGRATION,
                      table_name: :users,
                      column_name: :id,
                      job_arguments: []
                    )
                  end
                end
              RUBY
            end
          end

          context 'when the migration is no-op' do
            it 'does not throw offense' do
              expect_no_offenses(<<~RUBY)
                class FinalizeMyMigration < Gitlab::Database::Migration[2.1]
                  MIGRATION = 'MyMigration'

                  def up
                    # no-op
                  end
                end
              RUBY
            end
          end
        end
      end
    end
  end

  describe '#external_dependency_checksum' do
    it 'uses the Utils::BatchedBackgroundMigrationsDictionary.checksum' do
      allow(Gitlab::Utils::BatchedBackgroundMigrationsDictionary)
        .to receive(:checksum).and_return('aaaaa')

      expect(cop.external_dependency_checksum).to eq('aaaaa')
    end
  end
end
